# coding: utf-8

# Asynchronous Music Player Daemon client library for Python

# Copyright (C) 2015 Itaï BEN YAACOV

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import inspect

from . import _request, error


def worker(f):
    '''
    Decorator for worker functions.

    The first argument of a function decorated by @worker ('self', when decorating a method) must be either
    a. A Client.
    b. A WorkerGroup.
    c. An object with a WorkerGroup attribute named 'ampd_worker_group'.

    A worker function returns a Worker.
    '''

    argspec = inspect.getargspec(f)
    g = eval('lambda {0}: Worker(f({0}), {1}.ampd_worker_group)'.format(inspect.formatargspec(*argspec[:3])[1:-1], argspec.args[0]), {'Worker': Worker, 'f': f})
    g.__defaults__ = f.__defaults__
    g.__name__ = f.__name__
    g.__doc__ = f.__doc__
    return g


class Worker(object):
    '''
    AMPD worker.

    Created by a worker function, namely, a python generator decorated by @worker:

       @worker
       def worker_example(arg1, ...):
           ...
           reply = yield request1(a, b)
           ...
           reply = yield request2()
           ...

    The first argument for the worker function (self, if a method) must be one of:

    a. A Client.
    b. A WorkerGroup.
    c. An object with a WorkerGroup attribute named 'ampd_worker_group'.

    The function returns a new Worker immediately.
    Its code is later executed by the main loop.
    A statement of the form:

        reply = yield request

    suspends execution until something happens and a reply is available.
    The request can be:

    a. An MPD command (other than 'idle' and 'noidle').
       Returns when the server's reply arrives:

           yield play(5)
           reply = yield status()

    b. An idle request, emulating MPD's 'idle' command, with some improvements.
       Returns as soon as one of the conditions is satisfied, with a list of the satisfied conditions:

           reply = yield idle(condition, ... | iterable)
           reply = yield condition, ... | iterable

       Conditions can be:

       - An MPD SUBSYSTEM name (in uppercase).
       - ANY - match any subsystem.
       - TIMEOUT(ms) - satisfied after 'ms' milliseconds.
       - CONNECT - client is connected to server.
       - IDLE - client is idle.
       - WORKER(worker, ... | iterable) - all workers are done.

    c. Special request:

           yield _self()

       Returns the executing Worker.

    d. A request list.
       Emulates MPD's command list, returns a list of replies:

           yield request_list(request, ... | iterable)
           yield request, ... | iterable

    For MPD commands and subsystems see http://www.musicpd.org/doc/protocol/command_reference.html
    '''

    def __init__(self, gen, group):
        self._gen = gen
        self.name = gen.gi_code.co_name
        self._group = group
        self._atexit = []
        if group.workers == None:
            raise RuntimeError('Attempting to start a Worker in a closed WorkerGroup')
        group.workers.append(self)
        group.client._execute(self)

    def _close(self, close_generator=True):
        if self._gen:
            if close_generator:
                self._gen.close()
            self._group.workers.remove(self)
            for f in self._atexit:
                f()
            self._gen = self._group = self._atexit = None
            self.name += ' (closed)'

    def kill(self):
        '''
        Stop execution.
        '''
        if self._group:
            self._group.client._kill_worker(self)

    def add_atexit(self, f):
        '''
        Add a cleanup function to be called when the Worker exits.
        '''
        self._atexit.append(f)

    def __repr__(self):
        return self.name


class WorkerGroup(object):
    def __init__(self, client):
        self.client = client
        self.workers = []
        client._worker_groups.append(self)

    def close(self):
        if not self.client:
            return
        self.client._worker_groups.remove(self)
        while self.workers:
            self.client._kill_worker(self.workers[0])
        self.client = None

    @property
    def ampd_worker_group(self):
        return self

    @worker
    def connect_loop(self, connect_cb=None, disconnect_cb=None, connect_failed_cb=None):
        '''
        Convenience function for hooking into connect and disconnect events.
        Calls connect_cb() upon connect, disconnect_cb() upon disconnect, and connect_failed_cb() upon failed connect.
        '''
        while True:
            try:
                yield _request.conditions['CONNECT']
            except error.ConnectionError:
                connect_failed_cb and connect_failed_cb()
                continue
            connect_cb and connect_cb()
            try:
                yield _request.commands['idle']()
            except error.ConnectionError:
                disconnect_cb and disconnect_cb()

    def get_is_connected(self):
        return self.client.is_connected

    def get_protocol_version(self):
        return self.client.protocol_version



# coding: utf-8

# Asynchronous Music Player Daemon client library for Python

# Copyright (C) 2015 Itaï BEN YAACOV

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GLib, Gio
import socket


class Poller(object):
    def __init__(self, host, port, handle_read, handle_write, handle_error):
        self.host = host
        self.port = port
        self.handle_read = handle_read
        self.handle_write = handle_write
        self.handle_error = handle_error

    def close(self):
        pass


class PollerGLib(Poller):
    def __init__(self, *args, **kwargs):
        super(PollerGLib, self).__init__(*args, **kwargs)
        self.tcp_connection = self.socket = self.fd = None
        self.write_tag = self.other_tag = None
        self.cancel_connect = Gio.Cancellable.new()
        Gio.SocketClient.new().connect_to_host_async(self.host, self.port, self.cancel_connect, self.async_ready_cb)

    def close(self):
        super(PollerGLib, self).close()
        if self.cancel_connect:
            self.cancel_connect.cancel()
            self.cancel_connect = None
        if self.write_tag != None:
            GLib.source_remove(self.write_tag)
        if self.other_tag != None:
            GLib.source_remove(self.other_tag)
        if self.tcp_connection:
            self.tcp_connection.close()
        self.tcp_connection = self.socket = self.fd = None
        self.write_tag = self.other_tag = None

    def async_ready_cb(self, socket_client, task):
        self.cancel_connect = None
        try:
            self.tcp_connection = socket_client.connect_to_host_finish(task)
        except GLib.Error as error:
            self.handle_error(error.message)
            return
        self.socket = self.tcp_connection.get_socket()
        self.socket.set_blocking(False)
        self.fd = self.socket.get_fd()
        self.sock = socket.fromfd(self.fd, 0, 0)
        self.other_tag = GLib.io_add_watch(self.fd, GLib.IO_IN | GLib.IO_ERR | GLib.IO_HUP, self.callback_other)

    def callback_write(self, fd, condition):
        if self.handle_write(self.sock):
            return True
        else:
            self.write_tag = None
            return False

    def callback_other(self, fd, condition):
        if not condition & (GLib.IO_HUP | GLib.IO_ERR):
            if self.handle_read(self.sock):
                return True
        self.handle_error()
        self.other_tag = None
        return False

    def start_writing(self):
        if self.write_tag == None:
            self.write_tag = GLib.io_add_watch(self.fd, GLib.IO_OUT, self.callback_write)


class Scheduler(object):
    @staticmethod
    def connect_and_poll(host, port, handle_read, handle_write, handle_error):
        raise NotImplementedError

    @staticmethod
    def add_timeout(timeout, callback, parameter):
        raise NotImplementedError

    @staticmethod
    def remove_timeout(handler):
        raise NotImplementedError


class GLibScheduler(Scheduler):
    connect_and_poll = staticmethod(PollerGLib)
    add_timeout = staticmethod(GLib.timeout_add)
    remove_timeout = staticmethod(GLib.source_remove)



# coding: utf-8

# Asynchronous Music Player Daemon client library for Python

# Copyright (C) 2015 Itaï BEN YAACOV

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from .error import *


class Parser(object):
    WELCOME_PREFIX = 'OK MPD '
    SUCCESS = 'OK'
    ERROR_PREFIX = 'ACK'

    def process_line(self, line):
        if line.startswith(self.ERROR_PREFIX):
            return ReplyError(line)
        else:
            return self._parse_line(line)

class ParserWelcome(Parser):
    def _parse_line(self, line):
        if line.startswith(self.WELCOME_PREFIX):
            return line[len(self.WELCOME_PREFIX):]
        else:
            raise ProtocolError(line)

class ParserCommandList(Parser):
    LIST_SUCCESS = 'list_OK'

    def __init__(self, parsers):
        self.reply = []
        self.parsers = list(parsers)
        for parser in parsers:
            parser.SUCCESS = self.LIST_SUCCESS
        self.current = 0

    def process_line(self, line):
        if self.current == len(self.parsers):
            if line == self.SUCCESS:
                return self.reply
            raise ProtocolError(line)
        reply = self.parsers[self.current].process_line(line)
        if reply != None:
            self.reply.append(reply)
            self.current += 1

class ParserRaw(Parser):
    def __init__(self):
        self.reply = []

    def get_reply(self):
        return self.reply

    def _parse_line(self, line):
        if line == self.SUCCESS:
            return self.get_reply()
        delim = line.find(': ')
        if delim == -1:
            raise ProtocolError(line)
        self.reply.append( (line[:delim].lower(), line[delim+2:]) )

class ParserOK(ParserRaw):
    def get_reply(self):
        if self.reply == []:
            return True
        else:
            raise ProtocolError(self.reply)

class ParserDict(ParserRaw):
    def get_reply(self):
        return dict(self.reply)

class ParserDicts(ParserRaw):
    new_items = ['file', 'directory', 'playlist']

    def get_reply(self):
        items = {}
        for field, value in self.reply:
            if field in self.new_items:
                item = {}
                items.setdefault(field, []).append(item)
            item[field] = value
        return items

class _ParserFieldDict(ParserRaw):
    def __init__(self, field):
        super(_ParserFieldDict, self).__init__()
        self.field = field

    def get_reply(self):
        items = []
        for field, value in super(_ParserFieldDict, self).get_reply():
            if field == self.field:
                item = {}
                items.append(item)
            item[field] = value
        return items

def ParserFieldDict(field):
    return lambda: _ParserFieldDict(field)

class ParserList(ParserRaw):
    def get_reply(self):
        return [value for field, value in self.reply]

class ParserLists(ParserRaw):
    def get_reply(self):
        lists = {}
        for field, value in self.reply:
            lists.setdefault(field, []).append(value)
        return lists

class ParserItem(ParserRaw):
    def get_reply(self):
        reply = super(ParserItem, self).get_reply()
        if len(reply) == 1:
            return reply[0][1]
        else:
            raise ProtocolError(reply)

class _ParserSticker(ParserRaw):
    def get_reply(self):
        return [(field, value.split('=', 1)[1]) if field == 'sticker' else (field, value) for field, value in super(_ParserSticker, self).get_reply()]

class ParserStickerDict(_ParserFieldDict, _ParserSticker):
    def __init__(self):
        _ParserFieldDict.__init__(self, 'file')

    def get_reply(self):
        return {entry['file']: entry['sticker'] for entry in super(ParserStickerDict, self).get_reply()}

class ParserStickerItem(ParserItem, _ParserSticker):
    pass

class ParserStickerList(ParserRaw):
    def get_reply(self):
        return dict(value.split('=', 1) for field, value in super(ParserStickerList, self).get_reply())

class DeprecatedParser(Parser):
    def __init__(self, parser):
        self.parser = parser

    def process_line(self, line):
        reply = self.parser.process_line(line)
        if reply:
            return reply if isinstance(reply, ReplyError) else DeprecationWarning(reply)

def deprecate(parser_class):
    return lambda: DeprecatedParser(parser_class())



# coding: utf-8

# Asynchronous Music Player Daemon client library for Python

# Copyright (C) 2015 Itaï BEN YAACOV

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


'''
Asynchronous MPD client library

Uses the GLib main event loop and python generators for asynchronous communication with a Music Player Daemon server.

A Client connects to an MPD server.
MPD commands are executed by a Worker (or several), which is created by calling a worker function and is executed by the main loop.
A Worker can be started either directly within the Client, or, for more complex applications, within a WorkerGroup.
'''


import os
import logging


__version__ = '0.1.1b0'


_logger = logging.getLogger('ampd')
_logger.setLevel(os.environ.get('LOGLEVEL_AMPD', os.environ.get('LOGLEVEL', 'INFO')))
_handler = logging.StreamHandler()
_handler.setFormatter(logging.Formatter(fmt='%(levelname)s: %(name)s: %(message)s (%(pathname)s %(lineno)d)'))
_logger.addHandler(_handler)


__all__ = []

from ._client import Client
from ._worker import worker, Worker, WorkerGroup
__all__ += ['Client', 'worker', 'Worker', 'WorkerGroup']

from ._request import commands as commands_, conditions
globals().update(commands_)
globals().update(conditions)

__all__ += ['commands_', 'conditions']
__all__ += commands_.keys()
__all__ += conditions.keys()



# coding: utf-8

# Asynchronous Music Player Daemon client library for Python

# Copyright (C) 2015 Itaï BEN YAACOV

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class Error(Exception):
    def __repr__(self):
        return self.__class__.__name__ + ': ' + super(Error, self).__repr__()

    def __str__(self):
        return self.__class__.__name__ + ': ' + super(Error, self).__str__()

class ConnectionError(Error):
    pass

class ReplyError(Error):
    pass

class ProtocolError(Error):
    pass

class CommandError(Error):
    pass

class DeprecationWarning(Warning):
    pass



# coding: utf-8

# Asynchronous Music Player Daemon client library for Python

# Copyright (C) 2015 Itaï BEN YAACOV

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys

from .error import *
from . import scheduler, _request, _worker, _logger


_DEFAULT_CODING = None if sys.version_info >= (3, 0) else 'utf-8'

class Client(object):
    '''
    AMPD client object

    Establishes connection with the MPD server.

    Keeps track of:
    - The worker groups.
    - The active queue: requests send to server and waiting for a reply.
    - The passive list: 'idle' commands waiting for some conditions to hold.
    - The execution queue: workers for which a reply has arrived.
    '''

    def __init__(self, scheduler_=scheduler.GLibScheduler, coding=_DEFAULT_CODING, coding_server='utf-8', excepthook=sys.excepthook):
        '''
        Initialize a client.

        excepthook - override sys.excepthook for exceptions raised in workers.
        '''
        self._scheduler = scheduler_
        self._coding = coding
        self._coding_server = coding_server
        self._excepthook = excepthook
        self._passive_list = []
        self._execute_queue = []
        self._execute_timeout = None
        self._worker_groups = []
        self._poller = None
        self.is_connected = False
        self.protocol_version = None
        self.ampd_worker_group = _worker.WorkerGroup(self)

    def close(self):
        '''
        Close all workers and worker groups, disconnect from server.
        '''
        while self._worker_groups:
            self._worker_groups[0].close()
        self.disconnect()
        if self._execute_timeout:
            self._scheduler.remove_timeout(self._execute_timeout)
            self._run_execute_queue()

    def connect(self, host=None, port=6600, password=None):
        '''
        Connect to server.

        host     - '[password@]hostname[:port]'.  Default to $MPD_HOST or 'localhost'.
        port     - Ignored if given in the 'host' argument.
        password - Ignored if given in the 'host' argument.
        '''

        if not host:
            host = os.environ.get('MPD_HOST', 'localhost')
        if ':' in host:
            host, port = host.split(':', 1)
        if '@' in host:
            password, host = host.split('@', 1)

        if self._poller:
            self.disconnect()

        self._buff_in = self._buff_out = b''
        self._is_idle = False
        self._active_queue = []
        self._connect_worker()
        self._poller = self._scheduler.connect_and_poll(host, port, self._handle_read, self._handle_write, self._handle_error)
        self.password = password

    def disconnect(self):
        '''
        Disconnect from server.
        '''
        if self._poller:
            try:
                self._poller.sock.send('close')
            except:
                pass
            self._poller.close()
            self._poller = None
            self.is_connected = False
            self.protocol_version = None
            requests = self._active_queue + self._passive_list
            del self._buff_in, self._buff_out, self._active_queue
        else:
            requests = self._passive_list

        # Move everything to the execute queue with ConnectionError() reply
        self._execute_queue = [(worker, ConnectionError()) for worker, reply in self._execute_queue]
        for request in requests:
            request._execute(ConnectionError())

    def connect_loop(self, connect_cb=None, disconnect_cb=None, connect_failed_cb=None):
        return self.ampd_worker_group.connect_loop(connect_cb, disconnect_cb, connect_failed_cb)

    connect_loop.__doc__ = _worker.WorkerGroup.connect_loop.__doc__

    def _execute(self, worker, reply=None):
        self._execute_queue.append((worker, reply))
        if not self._execute_timeout:
            self._execute_timeout = self._scheduler.add_timeout(0, self._run_execute_queue)

    def _run_execute_queue(self):
        while self._execute_queue:
            worker, reply = self._execute_queue.pop(0)
            if not worker._gen:
                continue
            try:
                try:
                    request = worker._gen.throw(reply) if isinstance(reply, Exception) else worker._gen.send(reply)
                except:
                    del reply # If we threw an exception and got it back, python3 creates a reference cycle with the traceback.
                    worker._close()
                    raise
            except (StopIteration, ConnectionError):
                continue
            except:
                self._excepthook(*sys.exc_info())
                continue
            try:
                request = _request.Request.new(request)
                request._setup(worker, self)
            except Exception as reply:
                self._execute(worker, reply)

        self._execute_timeout = None

        if self.is_connected and not self._is_idle and not self._active_queue:
            self._idle_worker()

    def _kill_worker(self, worker):
        worker._close()
        try:
            self._execute_queue.remove(worker)
        except:
            for request in self._passive_list:
                if request.worker == worker:
                    request._execute(None)
                    break

    def _handle_error(self, message=None):
        self.disconnect()

    def _handle_write(self, sock):
        n = sock.send(self._buff_out)
        _logger.debug('Write: {}'.format(self._buff_out[:n]))
        self._buff_out = self._buff_out[n:]
        return len(self._buff_out) > 0

    def _handle_read(self, sock):
        data = sock.recv(10000)
        _logger.debug('Read: {}'.format(data))
        if not data:
            return False
        lines = (self._buff_in + data).split(b'\n')
        for line in lines[:-1]:
            line = line.decode(self._coding_server)
            if self._coding:
                line = line.encode(self._coding)
            if self._active_queue:
                self._active_queue[0]._process_line(line)
            else:
                raise ProtocolError('Unexpected data: ', line)
        self._buff_in = lines[-1]
        return True

    @_worker.worker
    def _connect_worker(self):
        self.protocol_version = yield _request.RequestWelcome()
        self.is_connected = True
        if self.password:
            yield _request.commands['password'](self.password)
        for request in list(self._passive_list):
            request._test(['connect'])

    @_worker.worker
    def _idle_worker(self):
        for request in self._passive_list:
            if request._test(['idle']):
                return
        _logger.debug('Going idle')
        subsystems = yield _request.RequestCommandIdle()
        for request in list(self._passive_list):
            request._test(subsystems)



# coding: utf-8

from setuptools import setup, find_packages
#import codecs
#import os

from ampd import __version__ as version
from ampd import __doc__ as long_description

#here = os.path.abspath(os.path.dirname(__file__))
#with codecs.open(os.path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
#    long_description = f.read()

setup(
    name='ampd',
    version=version,
    author='Itaï BEN YAACOV',
    author_email='candeb@free.fr',
    url='http://no.homepage.yet',
    #download_url='http://78.193.131.20/share/yBs6Rz8RgOo8zcfp/ampd',
    description='Asynchronous MPD client library',
    long_description=long_description,
    license='GPLv3+',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords=['mpd', 'Music Player Deamon', 'asynchronous'],
    packages=find_packages(),
)



# coding: utf-8

# Asynchronous Music Player Daemon client library for Python

# Copyright (C) 2015 Itaï BEN YAACOV

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from .error import *
from . import _parser, _logger


def sub(cls, name, **kwargs):
    kwargs['name'] = name
    return type(cls.__name__ + name.capitalize(), (cls,), kwargs)


class Condition(object):
    @staticmethod
    def __init__():
        raise NotImplementedError

    @classmethod
    def __repr__(cls):
        return cls.name.upper()

    @classmethod
    def test(cls, subsystems_):
        return [cls] if cls.name in subsystems_ else []

    @staticmethod
    def setup(request):
        return False

    @staticmethod
    def cleanup():
        pass


class ConditionSubsystem(Condition):
    pass

class ConditionAny(ConditionSubsystem):
    name = 'any'

    @staticmethod
    def test(subsystems_):
        return [subsystems[name.upper()] for name in subsystems_ if name.upper() in subsystems]


class ConditionTimeout(Condition):
    name = 'timeout'

    def __init__(self, timeout):
        self.timeout = int(timeout)
        self.timeout_handler = None

    def __del__(self):
        if self.timeout_handler:
            _logger.error('Unclean timeout handler')

    def __repr__(self):
        return '{}({})'.format(self.name.upper(), self.timeout)

    def setup(self, request):
        self.request = request
        self.timeout_handler = self.request.client._scheduler.add_timeout(self.timeout, self.timeout_cb)
        return False

    def cleanup(self):
        if self.timeout_handler:
            self.request.client._scheduler.remove_timeout(self.timeout_handler)
        self.timeout_handler = self.request = None

    def timeout_cb(self):
        self.timeout_handler = None
        self.request._execute([type(self)])


class ConditionConnect(Condition):
    name = 'connect'

    @staticmethod
    def setup(request):
        return request.client.is_connected


class ConditionIdle(Condition):
    name = 'idle'

    @staticmethod
    def setup(request):
        return request.client.is_connected and request.client._is_idle and not request.client._execute_queue


class ConditionWorker(Condition):
    name = 'worker'

    def __init__(self, worker):
        try:
            self.wait = list(worker)
        except:
            self.wait = [worker]

    def __repr__(self):
        return '{}({})'.format(self.name.upper(), self.wait)

    def setup(self, request):
        self.n = 0
        for worker in self.wait:
            if worker._gen:
                self.n += 1
                worker.add_atexit(self.worker_exit_cb)
        if not self.n:
            return True
        self.request = request
        return False

    def cleanup(self):
        self.request = None
        for worker in self.wait:
            try:
                worker._atexit.remove(self.worker_exit_cb)
            except:
                pass

    def worker_exit_cb(self):
        self.n -= 1
        if not self.n:
            self.request._execute([type(self)] + [self.wait])


subsystem_names = [
    'database',
    'update',
    'stored_playlist',
    'playlist',
    'player',
    'mixer',
    'output',
    'options',
    'sticker',
    'subscription',
    'message',
]

subsystems = {name.upper(): sub(ConditionSubsystem, name) for name in subsystem_names}
conditions = dict(subsystems)
for condition in (ConditionAny, ConditionTimeout, ConditionConnect, ConditionIdle, ConditionWorker):
    conditions[condition.name.upper()] = condition


class Request(object):
    def _setup(self, worker, client):
        self.worker = worker
        self.client = client

    def _execute(self, reply):
        _logger.debug('Request {!r}, reply {!r}'.format(self, reply))
        if self.worker._gen:
            self.client._execute(self.worker, reply)

    @staticmethod
    def new(request):
        if isinstance(request, Request):
            return request
        try:
            request = tuple(request)
        except:
            request = (request,)
        if all(isinstance(r, Condition) or (isinstance(r, type) and issubclass(r, Condition)) for r in request):
            return RequestPassive(*request)
        if len(request) != 1:
            return RequestList(*request)
        raise CommandError(*request)


class RequestActive(Request):
    def __init__(self):
        self.parser = self.parser_class()

    def __repr__(self):
        return self.commandline[:-1]

    def _setup(self, worker, client):
        super(RequestActive, self)._setup(worker, client)
        if client.is_connected:
            commandline = self.commandline
            if client._coding:
                commandline = commandline.decode(client._coding)
            if client._is_idle:
                commandline = 'noidle\n' + commandline
                client._is_idle = False
            client._buff_out += commandline.encode(client._coding_server)
            client._poller.start_writing()
            client._active_queue.append(self)
        else:
            self._execute(ConnectionError())

    def _process_line(self, line):
        reply = self.parser.process_line(line)
        if reply != None:
            del self.client._active_queue[0]
            self._execute(reply)


class RequestWelcome(RequestActive):
    parser_class = _parser.ParserWelcome

    def __init__(self):
        super(RequestWelcome, self).__init__()

    def __repr__(self):
        return 'WELCOME'

    def _setup(self, worker, client):
        Request._setup(self, worker, client)
        client._active_queue.append(self)


class RequestCommand(RequestActive):
    def __init__(self, *args):
        self.commandline = self.name + ''.join(' "{}"'.format(str(arg).replace('"', '\\"')) for arg in args) + '\n'
        super(RequestCommand, self).__init__()


class RequestCommandIdle(RequestActive):
    commandline = 'idle\n'
    parser_class = _parser.ParserList

    def _setup(self, worker, client):
        super(RequestCommandIdle, self)._setup(worker, client)
        client._is_idle = True

    def _execute(self, reply):
        self.client._is_idle = False
        super(RequestCommandIdle, self)._execute(reply)


class RequestPassive(Request):
    '''
    Emulates MPD's 'idle' command, with some improvements.
    Returns as soon as one of the conditions is satisfied, with a list of the satisfied conditions:

        reply = yield idle(condition, ... | iterable)
        reply = yield condition, ... | iterable

    Conditions can be:

    - An MPD SUBSYSTEM name (in uppercase).
    - ANY - match any subsystem.
    - TIMEOUT(ms) - satisfied after 'ms' milliseconds.
    - CONNECT - client is connected to server.
    - IDLE - client is idle.
    - WORKER(worker, ... | iterable) - all workers are done.
    '''
    def __init__(self, *args):
        self._conditions = list(args)

    def __repr__(self):
        return 'IDLE {!r}'.format(self._conditions)

    def _setup(self, worker, client):
        super(RequestPassive, self)._setup(worker, client)
        client._passive_list.append(self)
        reply = [condition for condition in self._conditions if condition.setup(self)]
        if reply:
            self._execute(reply)

    def _execute(self, reply):
        for condition in self._conditions:
            condition.cleanup()
        self.client._passive_list.remove(self)
        super(RequestPassive, self)._execute(reply)

    def _test(self, subsystems_):
        reply = sum(map(lambda c: c.test(subsystems_), self._conditions), [])
        if reply:
            self._execute(reply)
            return True
        else:
            return False


class RequestSelf(Request):
    '''
    Returns the Worker object associated with the executing worker:

        reply = yield _self()
    '''
    def _setup(self, worker, client):
        super(RequestSelf, self)._setup(worker, client)
        self._execute(worker)


class RequestList(Request):
    '''
    Emulates MPD's command list, returns a list of replies:

        reply = yield request_list(request, ... | iterable)
        reply = yield request, ... | iterable
    '''
    def __init__(self, *requests):
        if len(requests) == 1 and hasattr(requests[0], '__iter__'):
            requests = requests[0]
        self._requests = [self.new(request) for request in requests]
        self._replies = []

    def _setup(self, worker, client):
        super(RequestList, self)._setup(worker, client)
        self._worker_gen = self.worker._gen
        self.worker._gen = self._gen()
        self.client._execute(worker)

    def _execute(self, reply):
        self.worker._gen = self._worker_gen
        super(RequestList, self)._execute(reply)

    def _gen(self):
        for request in self._requests:
            try:
                reply = yield request
            except Exception as reply:
                self._execute(reply)
                return
            self._replies.append(reply)
        self._execute(self._replies)
        yield Request()

    def __repr__(self):
        return repr(self._requests)


command_parser = [
    ('add',                _parser.ParserOK),
    ('addid',              _parser.ParserItem),
    ('addtagid',           _parser.ParserRaw),
    ('channels',           _parser.ParserList),
    ('clear',              _parser.ParserOK),
    ('clearerror',         _parser.ParserOK),
    ('cleartagid',         _parser.ParserRaw),
    ('close',              lambda: None),
    ('commands',           _parser.ParserList),
    ('config',             _parser.ParserDict),
    ('consume',            _parser.ParserOK),
    ('count',              _parser.ParserDict),
    ('crossfade',          _parser.ParserOK),
    ('currentsong',        _parser.ParserDict),
    ('decoders',           _parser.ParserLists),
    ('delete',             _parser.ParserOK),
    ('deleteid',           _parser.ParserOK),
    ('disableoutput',      _parser.ParserOK),
    ('enableoutput',       _parser.ParserOK),
    ('find',               _parser.ParserFieldDict('file')),
    ('findadd',            _parser.ParserOK),
    ('kill',               lambda: None),
    ('list',               _parser.ParserList),
    ('listall',            _parser.ParserLists),
    ('listallinfo',        _parser.ParserDicts),
    ('listfiles',          _parser.ParserDicts),
    ('listmounts',         _parser.ParserFieldDict('mount')),
    ('listneighbors',      _parser.ParserFieldDict('neighbor')),
    ('listplaylist',       _parser.ParserList),
    ('listplaylistinfo',   _parser.ParserFieldDict('file')),
    ('listplaylists',      _parser.ParserFieldDict('playlist')),
    ('load',               _parser.ParserOK),
    ('lsinfo',             _parser.ParserDicts),
    ('mixrampdb',          _parser.ParserOK),
    ('mixrampdelay',       _parser.ParserOK),
    ('mount',              _parser.ParserRaw),
    ('move',               _parser.ParserOK),
    ('moveid',             _parser.ParserOK),
    ('next',               _parser.ParserOK),
    ('notcommands',        _parser.ParserList),
    ('outputs',            _parser.ParserFieldDict('outputid')),
    ('password',           _parser.ParserOK),
    ('pause',              _parser.ParserOK),
    ('ping',               _parser.ParserOK),
    ('play',               _parser.ParserOK),
    ('playid',             _parser.ParserOK),
    ('playlist',           _parser.ParserList),
    ('playlistadd',        _parser.ParserOK),
    ('playlistclear',      _parser.ParserOK),
    ('playlistdelete',     _parser.ParserOK),
    ('playlistfind',       _parser.ParserFieldDict('file')),
    ('playlistid',         _parser.ParserFieldDict('file')),
    ('playlistinfo',       _parser.ParserFieldDict('file')),
    ('playlistmove',       _parser.ParserOK),
    ('playlistsearch',     _parser.ParserFieldDict('file')),
    ('plchanges',          _parser.ParserFieldDict('file')),
    ('plchangesposid',     _parser.ParserFieldDict('cpos')),
    ('previous',           _parser.ParserOK),
    ('prio',               _parser.ParserOK),
    ('prioid',             _parser.ParserOK),
    ('random',             _parser.ParserOK),
    ('rangeid',            _parser.ParserOK),
    ('readcomments',       _parser.ParserDict),
    ('readmessages',       _parser.ParserFieldDict('channel')),
    ('rename',             _parser.ParserOK),
    ('repeat',             _parser.ParserOK),
    ('replay_gain_mode',   _parser.ParserOK),
    ('replay_gain_status', _parser.ParserItem),
    ('rescan',             _parser.ParserItem),
    ('rm',                 _parser.ParserOK),
    ('save',               _parser.ParserOK),
    ('search',             _parser.ParserFieldDict('file')),
    ('searchadd',          _parser.ParserOK),
    ('searchaddpl',        _parser.ParserOK),
    ('seek',               _parser.ParserOK),
    ('seekcur',            _parser.ParserOK),
    ('seekid',             _parser.ParserOK),
    ('sendmessage',        _parser.ParserOK),
    ('setvol',             _parser.ParserOK),
    ('shuffle',            _parser.ParserOK),
    ('single',             _parser.ParserOK),
    ('stats',              _parser.ParserDict),
    ('status',             _parser.ParserDict),
    ('sticker_get',        _parser.ParserStickerItem),
    ('sticker_set',        _parser.ParserOK),
    ('sticker_delete',     _parser.ParserOK),
    ('sticker_list',       _parser.ParserStickerList),
    ('sticker_find',       _parser.ParserStickerDict),
    ('stop',               _parser.ParserOK),
    ('subscribe',          _parser.ParserOK),
    ('swap',               _parser.ParserOK),
    ('swapid',             _parser.ParserOK),
    ('tagtypes',           _parser.ParserList),
    ('toggleoutput',       _parser.ParserOK),
    ('unmount',            _parser.ParserRaw),
    ('unsubscribe',        _parser.ParserOK),
    ('update',             _parser.ParserItem),
    ('urlhandlers',        _parser.ParserList),
    ('volume',             _parser.deprecate(_parser.ParserOK)),
]

commands = {command: sub(RequestCommand, command, parser_class=staticmethod(parser_class)) for command, parser_class in command_parser}
commands['idle'] = RequestPassive
commands['_self'] = RequestSelf
commands['request_list'] = RequestList
